#   Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Customized losses."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


class BatchSoftmax(tf.keras.losses.Loss):
  """Compute batch softmax over batch similarities.

  This softmax loss takes in-batch negatives without considering
  negatives out of batch.
  """

  def __init__(self, name='batch_softmax', **kwargs):
    super(BatchSoftmax, self).__init__(name=name, **kwargs)

  @tf.function
  def call(self, y_true: tf.Tensor, y_pred: tf.Tensor):
    """Compute in batch softmax loss.

    Args:
      y_true: the true labels with shape [batch_size, 1].
      y_pred: pre-calculated similarities matrix between context embedding and
        full vocab label embeddings, y_pred is expected to be the similarities
        with shape [batch_size,label_embedding_vocab_size].

    Returns:
      The softmax loss with in-batch negatives.
    """
    logits = tf.keras.backend.cast(y_pred, 'float32')
    logits = tf.gather(y_pred, tf.transpose(y_true)[0], axis=1)
    full_labels = tf.eye(
        tf.shape(logits)[0], tf.shape(logits)[1], dtype=tf.dtypes.float32)
    batch_loss = tf.nn.softmax_cross_entropy_with_logits(
        labels=full_labels, logits=logits)
    loss = tf.reduce_mean(batch_loss)
    return loss


class GlobalSoftmax(tf.keras.losses.Loss):
  """Compute softmax over similarities.

  This loss fuction computes softmax over similarities between context and
  full vocab label embeddings, considering full label vocab non-label
  predictions as negatives. This is currently the default loss in the model.
  """

  def __init__(self, name='global_softmax', **kwargs):
    super(GlobalSoftmax, self).__init__(name=name, **kwargs)

  @tf.function
  def call(self, y_true: tf.Tensor, y_pred: tf.Tensor):
    """Compute softmax loss with full vocab labels as negatives.

    Args:
      y_true: the true labels with shape [batch_size, 1].
      y_pred: the pre-calculated similarities matrix with shape [batch_size,
        label_embedding_vocab_size]

    Returns:
      The softmax loss with full vocab labels as negatives.
    """
    logits = tf.keras.backend.cast(y_pred, 'float32')
    # Compose a metric to tell which column represents the similarity between
    # the example and the label. For each row, only the item at index of the
    # label of that example will be set as 1.
    full_labels = tf.one_hot(
        tf.transpose(y_true)[0], tf.shape(logits)[1], dtype=tf.dtypes.float32)
    batch_loss = tf.nn.softmax_cross_entropy_with_logits(
        labels=full_labels, logits=logits)
    loss = tf.reduce_mean(batch_loss)
    return loss



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Recommendation dataloader."""

import os
import tempfile
from unittest import mock

import tensorflow as tf

from tensorflow_examples.lite.model_maker.core.data_util import recommendation_config
from tensorflow_examples.lite.model_maker.third_party.recommendation.ml.data import example_generation_movielens as gen

MOVIE_SIZE = 101
RATING_SIZE = 5000
USER_SIZE = 50

TRAIN_SIZE = 4455
TEST_SIZE = 495
VOCAB_SIZE = 101
MAX_ITEM_ID = 999


def _generate_fake_data(data_dir):
  """Generates fake data to files.

  It generates 3 files.
  - movies.dat: movies data, with format per line:
               MovieID::Title::Genres.
  - users.dat: users data, with format per line:
               UserID::Gender::Age::Occupation::Zip-code.
  - ratings.dat: movie ratings by users, with format per line:
               UserID::MovieID::Rating::Timestamp
  It aligns with movielens dataset. IDs start from 1, and 0 is reserved for OOV.

  Args:
    data_dir: str, dir name to generate dataset.
  """
  if not tf.io.gfile.exists(data_dir):
    tf.io.gfile.makedirs(data_dir)

  # Movies:
  # MovieID::Title::Genres
  movies = ['{i}::title{i}::genre1|genre2'.format(i=i) for i in range(1, 101)]
  movies.append('999::title999::genere10')  # Add a movie with a larger id.
  movie_file = os.path.join(data_dir, 'movies.dat')
  _write_file_by_lines(movies, movie_file)

  # Users:
  # UserID::Gender::Age::Occupation::Zip-code
  users = ['{user}::F::0::0::00000'.format(user=user) for user in range(1, 51)]
  user_file = os.path.join(data_dir, 'users.dat')
  _write_file_by_lines(users, user_file)

  # Ratings:
  # UserID::MovieID::Rating::Timestamp
  ratings = []
  for user in range(1, 51):
    ratings += [
        '{user}::{movie}::5::{timestamp}'.format(
            user=user, movie=movie, timestamp=978000000 + 1)
        for movie in range(1, 101)
    ]
  rating_file = os.path.join(data_dir, 'ratings.dat')
  _write_file_by_lines(ratings, rating_file)


def _write_file_by_lines(data, filename):
  """Writes data to file line per line."""
  with tf.io.gfile.GFile(filename, 'w') as f:
    for line in data:
      f.write(str(line))
      f.write('\n')


def setup_fake_testdata(obj):
  """Setup fake testdata folder.

  This function creates new attrs:
  - test_tempdir: temporary dir (optional), if not exists.
  - download_dir: datasets dir for downloaded zip.
  - dataset_dir: extracted dir for movielens data.

  Args:
    obj: object, usually TestCase instance's self or cls.
  """
  if not hasattr(obj, 'test_tempdir'):
    obj.test_tempdir = tempfile.mkdtemp()
  obj.download_dir = os.path.join(obj.test_tempdir, 'download')
  obj.dataset_dir = os.path.join(obj.download_dir, 'fake_movielens')
  _generate_fake_data(obj.dataset_dir)


def patch_download_and_extract_data(data_dir):
  """Patch download and extract data for testing.

  The common usage is to generate data loader:

  with patch_download_and_extract_data(movielens_dir):
    train_loader = RecommendationDataLoader.from_movielens(
        generated_dir, 'train', test_tempdir)

  Args:
    data_dir: str, path to data dir.

  Returns:
    mocked context.
  """

  def side_effect(*args, **kwargs):
    del args, kwargs
    tf.compat.v1.logging.info('Use patched dataset dir: %s', data_dir)
    return data_dir

  return mock.patch.object(
      gen, 'download_and_extract_data', side_effect=side_effect)


def get_input_spec(encoder_type='cnn') -> recommendation_config.InputSpec:
  """Gets input spec (for test).

  Input spec defines how the input features are extracted.

  Args:
    encoder_type: str. Case-insensitive {'CNN', 'LSTM', 'BOW'}.

  Returns:
    InputSpec.
  """
  etype = encoder_type.upper()
  if etype not in {'CNN', 'LSTM', 'BOW'}:
    raise ValueError('Not support encoder_type: {}'.format(etype))

  return recommendation_config.InputSpec(
      activity_feature_groups=[
          # Group #1: defines how features are grouped in the first Group.
          dict(
              features=[
                  # First feature.
                  dict(
                      feature_name='context_movie_id',  # Feature name
                      feature_type='INT',  # Feature type
                      vocab_size=3953,  # ID size (number of IDs)
                      embedding_dim=8,  # Projected feature embedding dim
                      feature_length=10,  # History length of 10.
                  ),
                  # Maybe more features...
              ],
              encoder_type='CNN',  # CNN encoder (e.g. CNN, LSTM, BOW)
          ),
          # Maybe more groups...
      ],
      label_feature=dict(
          feature_name='label_movie_id',  # Label feature name
          feature_type='INT',  # Label type
          vocab_size=3953,  # Label size (number of classes)
          embedding_dim=8,  # Label embedding demension
          feature_length=1,  # Exactly 1 label
      ),
  )


def get_model_hparams() -> recommendation_config.ModelHParams:
  """Gets model hparams (for test).

  ModelHParams defines the model architecture.

  Returns:
    ModelHParams.
  """
  return recommendation_config.ModelHParams(
      hidden_layer_dims=[32, 32],  # Hidden layers dimension.
      eval_top_k=[1, 5],  # Eval top 1 and top 5.
      conv_num_filter_ratios=[2, 4],  # For CNN encoder, conv filter mutipler.
      conv_kernel_size=16,  # For CNN encoder, base kernel size.
      lstm_num_units=16,  # For LSTM/RNN, num units.
      num_predictions=10,  # Number of output predictions. Select top 10.
  )



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MetadataLoader to load metadata for each input data."""

import enum
import os
from typing import AnyStr, Callable

from tensorflow_examples.lite.model_maker.core.api.api_util import mm_export


@mm_export("searcher.MetadataType")
@enum.unique
class MetadataType(enum.Enum):
  FROM_FILE_NAME = 1
  FROM_DAT_FILE = 2


class MetadataLoader(object):
  """MetadataLoader class to load metadata for each input data."""

  def __init__(self, func: Callable[..., AnyStr]) -> None:
    self._func = func

  def load(self, path: str, **kwargs) -> AnyStr:
    """Loads metadata from input data path."""
    return self._func(path, **kwargs)

  @classmethod
  def from_file_name(cls) -> "MetadataLoader":
    """Loads the file name as metadata."""

    def _load_metadata(path):
      return os.path.splitext(os.path.basename(path))[0]

    return cls(_load_metadata)

  @classmethod
  def from_dat_file(cls) -> "MetadataLoader":
    """Loads the metadata from .dat file.

    Reads metadata from a file assumed to be stored next to the (image, audio or
    text) input file, with the same basename and a different extension ".dat",
    e.g: if input file is "/path/to/foo.jpg", read metadata from
    "/path/to/foo.dat". When loading metadata, user can pass `mode` parameter
    which is a string that defines which mode you want to open the file in: "r"
    (text metadata) and "rb" (binary metadata) are supported.

    Returns:
      MetadataLoader object.
    """

    def _load_metadata(path, mode="r"):
      if mode != "r" and mode != "rb":
        raise ValueError(
            "`mode` must be \"r\" or \"rb\" to read text or binary metadata.")

      metadata_file_path = os.path.splitext(path)[0] + ".dat"
      with open(metadata_file_path, mode) as f:
        return f.read()

    return cls(_load_metadata)



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""The main training script."""
import multiprocessing
import os
from absl import app
from absl import flags
from absl import logging
import numpy as np
import tensorflow.compat.v1 as tf
from tensorflow.compat.v1 import estimator as tf_estimator

from tensorflow_examples.lite.model_maker.third_party.efficientdet import dataloader
from tensorflow_examples.lite.model_maker.third_party.efficientdet import det_model_fn
from tensorflow_examples.lite.model_maker.third_party.efficientdet import hparams_config
from tensorflow_examples.lite.model_maker.third_party.efficientdet import utils

flags.DEFINE_string(
    'tpu',
    default=None,
    help='The Cloud TPU to use for training. This should be either the name '
    'used when creating the Cloud TPU, or a grpc://ip.address.of.tpu:8470 '
    'url.')
flags.DEFINE_string(
    'gcp_project',
    default=None,
    help='Project name for the Cloud TPU-enabled project. If not specified, we '
    'will attempt to automatically detect the GCE project from metadata.')
flags.DEFINE_string(
    'tpu_zone',
    default=None,
    help='GCE zone where the Cloud TPU is located in. If not specified, we '
    'will attempt to automatically detect the GCE project from metadata.')
flags.DEFINE_string('eval_name', default=None, help='Eval job name')
flags.DEFINE_enum('strategy', None, ['tpu', 'gpus', ''],
                  'Training: gpus for multi-gpu, if None, use TF default.')

flags.DEFINE_bool('use_fake_data', False, 'Use fake input.')
flags.DEFINE_bool(
    'use_xla', False,
    'Use XLA even if strategy is not tpu. If strategy is tpu, always use XLA, '
    'and this flag has no effect.')
flags.DEFINE_string('model_dir', None, 'Location of model_dir')
flags.DEFINE_string(
    'backbone_ckpt', '', 'Location of the ResNet50 checkpoint to use for model '
    'initialization.')
flags.DEFINE_string('ckpt', None,
                    'Start training from this EfficientDet checkpoint.')

flags.DEFINE_string(
    'hparams', '', 'Comma separated k=v pairs of hyperparameters or a module'
    ' containing attributes to use as hyperparameters.')
flags.DEFINE_integer(
    'num_cores', default=8, help='Number of TPU cores for training')
flags.DEFINE_bool('use_spatial_partition', False, 'Use spatial partition.')
flags.DEFINE_integer(
    'num_cores_per_replica',
    default=2,
    help='Number of TPU cores per replica when using spatial partition.')
flags.DEFINE_multi_integer(
    'input_partition_dims', [1, 2, 1, 1],
    'A list that describes the partition dims for all the tensors.')
flags.DEFINE_integer('train_batch_size', 64, 'global training batch size')
flags.DEFINE_integer('eval_batch_size', 1, 'global evaluation batch size')
flags.DEFINE_integer('eval_samples', 5000, 'Number of samples for eval.')
flags.DEFINE_integer('iterations_per_loop', 1000,
                     'Number of iterations per TPU training loop')
flags.DEFINE_integer('save_checkpoints_steps', 1000,
                     'Number of iterations per checkpoint save')
flags.DEFINE_string(
    'train_file_pattern', None,
    'Glob for training data files (e.g., COCO train - minival set)')
flags.DEFINE_string('val_file_pattern', None,
                    'Glob for evaluation tfrecords (e.g., COCO val2017 set)')
flags.DEFINE_string(
    'val_json_file', None,
    'COCO validation JSON containing golden bounding boxes. If None, use the '
    'ground truth from the dataloader. Ignored if testdev_dir is not None.')
flags.DEFINE_string('testdev_dir', None,
                    'COCO testdev dir. If not None, ignorer val_json_file.')
flags.DEFINE_integer('num_examples_per_epoch', 120000,
                     'Number of examples in one epoch')
flags.DEFINE_integer('num_epochs', None, 'Number of epochs for training')
flags.DEFINE_string('mode', 'train',
                    'Mode to run: train or eval (default: train)')
flags.DEFINE_string('model_name', 'efficientdet-d1', 'Model name.')
flags.DEFINE_bool('eval_after_train', False, 'Run one eval after the '
                  'training finishes.')
flags.DEFINE_bool('profile', False, 'Profile training performance.')
flags.DEFINE_integer(
    'tf_random_seed', None, 'Sets the TF graph seed for deterministic execution'
    ' across runs (for debugging).')

# For Eval mode
flags.DEFINE_integer('min_eval_interval', 180,
                     'Minimum seconds between evaluations.')
flags.DEFINE_integer(
    'eval_timeout', None,
    'Maximum seconds between checkpoints before evaluation terminates.')

# for train_and_eval mode
flags.DEFINE_bool(
    'run_epoch_in_child_process', False,
    'This option helps to rectify CPU memory leak. If True, every epoch is '
    'run in a separate process for train and eval and memory will be cleared.'
    'Drawback: need to kill 2 processes if trainining needs to be interrupted.')

FLAGS = flags.FLAGS


def main(_):
  if FLAGS.strategy == 'tpu':
    tf.disable_eager_execution()
    tpu_cluster_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(
        FLAGS.tpu, zone=FLAGS.tpu_zone, project=FLAGS.gcp_project)
    tpu_grpc_url = tpu_cluster_resolver.get_master()
    tf.Session.reset(tpu_grpc_url)
  else:
    tpu_cluster_resolver = None

  # Check data path
  if FLAGS.mode in ('train', 'train_and_eval'):
    if FLAGS.train_file_pattern is None:
      raise RuntimeError('Must specify --train_file_pattern for train.')
  if FLAGS.mode in ('eval', 'train_and_eval'):
    if FLAGS.val_file_pattern is None:
      raise RuntimeError('Must specify --val_file_pattern for eval.')

  # Parse and override hparams
  config = hparams_config.get_detection_config(FLAGS.model_name)
  config.override(FLAGS.hparams)
  if FLAGS.num_epochs:  # NOTE: remove this flag after updating all docs.
    config.num_epochs = FLAGS.num_epochs

  # Parse image size in case it is in string format.
  config.image_size = utils.parse_image_size(config.image_size)

  # The following is for spatial partitioning. `features` has one tensor while
  # `labels` had 4 + (`max_level` - `min_level` + 1) * 2 tensors. The input
  # partition is performed on `features` and all partitionable tensors of
  # `labels`, see the partition logic below.
  # In the TPUEstimator context, the meaning of `shard` and `replica` is the
  # same; follwing the API, here has mixed use of both.
  if FLAGS.use_spatial_partition:
    # Checks input_partition_dims agrees with num_cores_per_replica.
    if FLAGS.num_cores_per_replica != np.prod(FLAGS.input_partition_dims):
      raise RuntimeError('--num_cores_per_replica must be a product of array'
                         'elements in --input_partition_dims.')

    labels_partition_dims = {
        'mean_num_positives': None,
        'source_ids': None,
        'groundtruth_data': None,
        'image_scales': None,
        'image_masks': None,
    }
    # The Input Partition Logic: We partition only the partition-able tensors.
    feat_sizes = utils.get_feat_sizes(
        config.get('image_size'), config.get('max_level'))
    for level in range(config.get('min_level'), config.get('max_level') + 1):

      def _can_partition(spatial_dim):
        partitionable_index = np.where(
            spatial_dim % np.array(FLAGS.input_partition_dims) == 0)
        return len(partitionable_index[0]) == len(FLAGS.input_partition_dims)

      spatial_dim = feat_sizes[level]
      if _can_partition(spatial_dim['height']) and _can_partition(
          spatial_dim['width']):
        labels_partition_dims['box_targets_%d' %
                              level] = FLAGS.input_partition_dims
        labels_partition_dims['cls_targets_%d' %
                              level] = FLAGS.input_partition_dims
      else:
        labels_partition_dims['box_targets_%d' % level] = None
        labels_partition_dims['cls_targets_%d' % level] = None
    num_cores_per_replica = FLAGS.num_cores_per_replica
    input_partition_dims = [FLAGS.input_partition_dims, labels_partition_dims]
    num_shards = FLAGS.num_cores // num_cores_per_replica
  else:
    num_cores_per_replica = None
    input_partition_dims = None
    num_shards = FLAGS.num_cores

  params = dict(
      config.as_dict(),
      model_name=FLAGS.model_name,
      iterations_per_loop=FLAGS.iterations_per_loop,
      model_dir=FLAGS.model_dir,
      num_shards=num_shards,
      num_examples_per_epoch=FLAGS.num_examples_per_epoch,
      strategy=FLAGS.strategy,
      backbone_ckpt=FLAGS.backbone_ckpt,
      ckpt=FLAGS.ckpt,
      val_json_file=FLAGS.val_json_file,
      testdev_dir=FLAGS.testdev_dir,
      profile=FLAGS.profile,
      mode=FLAGS.mode)
  config_proto = tf.ConfigProto(
      allow_soft_placement=True, log_device_placement=False)
  if FLAGS.strategy != 'tpu':
    if FLAGS.use_xla:
      config_proto.graph_options.optimizer_options.global_jit_level = (
          tf.OptimizerOptions.ON_1)
      config_proto.gpu_options.allow_growth = True

  model_dir = FLAGS.model_dir
  model_fn_instance = det_model_fn.get_model_fn(FLAGS.model_name)
  max_instances_per_image = config.max_instances_per_image
  if FLAGS.eval_samples:
    eval_steps = int((FLAGS.eval_samples + FLAGS.eval_batch_size - 1) //
                     FLAGS.eval_batch_size)
  else:
    eval_steps = None
  total_examples = int(config.num_epochs * FLAGS.num_examples_per_epoch)
  train_steps = total_examples // FLAGS.train_batch_size
  logging.info(params)

  if not tf.io.gfile.exists(model_dir):
    tf.io.gfile.makedirs(model_dir)

  config_file = os.path.join(model_dir, 'config.yaml')
  if not tf.io.gfile.exists(config_file):
    tf.io.gfile.GFile(config_file, 'w').write(str(config))

  train_input_fn = dataloader.InputReader(
      FLAGS.train_file_pattern,
      is_training=True,
      use_fake_data=FLAGS.use_fake_data,
      max_instances_per_image=max_instances_per_image)
  eval_input_fn = dataloader.InputReader(
      FLAGS.val_file_pattern,
      is_training=False,
      use_fake_data=FLAGS.use_fake_data,
      max_instances_per_image=max_instances_per_image)

  if FLAGS.strategy == 'tpu':
    tpu_config = tf_estimator.tpu.TPUConfig(
        FLAGS.iterations_per_loop if FLAGS.strategy == 'tpu' else 1,
        num_cores_per_replica=num_cores_per_replica,
        input_partition_dims=input_partition_dims,
        per_host_input_for_training=tf_estimator.tpu.InputPipelineConfig
        .PER_HOST_V2)
    run_config = tf_estimator.tpu.RunConfig(
        cluster=tpu_cluster_resolver,
        model_dir=model_dir,
        log_step_count_steps=FLAGS.iterations_per_loop,
        session_config=config_proto,
        tpu_config=tpu_config,
        save_checkpoints_steps=FLAGS.save_checkpoints_steps,
        tf_random_seed=FLAGS.tf_random_seed,
    )
    # TPUEstimator can do both train and eval.
    train_est = tf_estimator.tpu.TPUEstimator(
        model_fn=model_fn_instance,
        train_batch_size=FLAGS.train_batch_size,
        eval_batch_size=FLAGS.eval_batch_size,
        config=run_config,
        params=params)
    eval_est = train_est
  else:
    strategy = None
    if FLAGS.strategy == 'gpus':
      strategy = tf.distribute.MirroredStrategy()
    run_config = tf_estimator.RunConfig(
        model_dir=model_dir,
        train_distribute=strategy,
        log_step_count_steps=FLAGS.iterations_per_loop,
        session_config=config_proto,
        save_checkpoints_steps=FLAGS.save_checkpoints_steps,
        tf_random_seed=FLAGS.tf_random_seed,
    )

    def get_estimator(global_batch_size):
      params['num_shards'] = getattr(strategy, 'num_replicas_in_sync', 1)
      params['batch_size'] = global_batch_size // params['num_shards']
      return tf_estimator.Estimator(
          model_fn=model_fn_instance, config=run_config, params=params)

    # train and eval need different estimator due to different batch size.
    train_est = get_estimator(FLAGS.train_batch_size)
    eval_est = get_estimator(FLAGS.eval_batch_size)

  # start train/eval flow.
  if FLAGS.mode == 'train':
    train_est.train(input_fn=train_input_fn, max_steps=train_steps)
    if FLAGS.eval_after_train:
      eval_est.evaluate(input_fn=eval_input_fn, steps=eval_steps)

  elif FLAGS.mode == 'eval':
    # Run evaluation when there's a new checkpoint
    for ckpt in tf.train.checkpoints_iterator(
        FLAGS.model_dir,
        min_interval_secs=FLAGS.min_eval_interval,
        timeout=FLAGS.eval_timeout):

      logging.info('Starting to evaluate.')
      try:
        eval_results = eval_est.evaluate(
            eval_input_fn, steps=eval_steps, name=FLAGS.eval_name)
        # Terminate eval job when final checkpoint is reached.
        try:
          current_step = int(os.path.basename(ckpt).split('-')[1])
        except IndexError:
          logging.info('%s has no global step info: stop!', ckpt)
          break

        utils.archive_ckpt(eval_results, eval_results['AP'], ckpt)
        if current_step >= train_steps:
          logging.info('Eval finished step %d/%d', current_step, train_steps)
          break

      except tf.errors.NotFoundError:
        # Checkpoint might be not already deleted by the time eval finished.
        # We simply skip ssuch case.
        logging.info('Checkpoint %s no longer exists, skipping.', ckpt)

  elif FLAGS.mode == 'train_and_eval':
    ckpt = tf.train.latest_checkpoint(FLAGS.model_dir)
    try:
      step = int(os.path.basename(ckpt).split('-')[1])
      current_epoch = (
          step * FLAGS.train_batch_size // FLAGS.num_examples_per_epoch)
      logging.info('found ckpt at step %d (epoch %d)', step, current_epoch)
    except (IndexError, TypeError):
      logging.info('Folder %s has no ckpt with valid step.', FLAGS.model_dir)
      current_epoch = 0

    def run_train_and_eval(e):
      print('\n   =====> Starting training, epoch: %d.' % e)
      train_est.train(
          input_fn=train_input_fn,
          max_steps=e * FLAGS.num_examples_per_epoch // FLAGS.train_batch_size)
      print('\n   =====> Starting evaluation, epoch: %d.' % e)
      eval_results = eval_est.evaluate(input_fn=eval_input_fn, steps=eval_steps)
      ckpt = tf.train.latest_checkpoint(FLAGS.model_dir)
      utils.archive_ckpt(eval_results, eval_results['AP'], ckpt)

    epochs_per_cycle = 1  # higher number has less graph construction overhead.
    for e in range(current_epoch + 1, config.num_epochs + 1, epochs_per_cycle):
      if FLAGS.run_epoch_in_child_process:
        p = multiprocessing.Process(target=run_train_and_eval, args=(e,))
        p.start()
        p.join()
        if p.exitcode != 0:
          return p.exitcode
      else:
        tf.compat.v1.reset_default_graph()
        run_train_and_eval(e)

  else:
    logging.info('Invalid mode: %s', FLAGS.mode)


if __name__ == '__main__':
  app.run(main)



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Writes metadata and tokenize file to the Bert QA models."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow_examples.lite.model_maker.core.task.metadata_writers.bert.metadata_writer_for_bert import bert_qa_inputs
from tensorflow_examples.lite.model_maker.core.task.metadata_writers.bert.metadata_writer_for_bert import MetadataPopulatorForBert
from tensorflow_examples.lite.model_maker.core.task.metadata_writers.bert.metadata_writer_for_bert import ModelSpecificInfo
from tensorflow_examples.lite.model_maker.core.task.metadata_writers.bert.metadata_writer_for_bert import Tokenizer
from tflite_support import metadata_schema_py_generated as _metadata_fb


def bert_qa_outputs(start_logits_name, end_logits_name):
  """Creates the output tensor names of a Bert question answerer model in order.

  The names correspond to `Tensor.name` in the TFLite schema. It helps to
  determine the tensor order when populating the metadata.

  Args:
    start_logits_name: name of the start logits tensor, which represents the
      start position of the answer span.
    end_logits_name: name of the end logits tensor, which represents the end
      position of the answer span.

  Returns:
    The output name list.
  """
  return [start_logits_name, end_logits_name]


class QuestionAnswererInfo(ModelSpecificInfo):
  """Holds information specificly tied to a Bert question answerer model."""

  def __init__(self,
               name,
               version,
               description,
               input_names,
               output_names,
               tokenizer_type,
               vocab_file=None,
               sp_model=None):
    """Constructor for ModelSpecificInfo.

    Args:
      name: name of the model in string.
      version: version of the model in string.
      description: description of the model.
      input_names: the name list returned by bert_qa_inputs.
      output_names: the name list returned by bert_qa_outputs.
      tokenizer_type: one of the tokenizer types in Tokenizer.
      vocab_file: the vocab file name to be packed into the model. If the
        tokenizer is BERT_TOKENIZER, the vocab file is required; if the
        tokenizer is SENTENCE_PIECE, the vocab file is optional.
      sp_model: the SentencePiece model file, only valid for the SENTENCE_PIECE
        tokenizer.
    """
    self.output_names = output_names
    super(QuestionAnswererInfo,
          self).__init__(name, version, description, input_names,
                         tokenizer_type, vocab_file, sp_model)


DEFAULT_DESCRIPTION = (
    "Answers questions based on the content of a given "
    "passage. To integrate the model into your app, try the "
    "`BertQuestionAnswerer` API in the TensorFlow Lite Task "
    "library. `BertQuestionAnswerer` takes a passage string and a "
    "query string, and returns the answer strings. It encapsulates "
    "the processing logic of inputs and outputs and runs the "
    "inference with the best practice.")

DEFAULT_INPUT_NAMES = bert_qa_inputs(
    ids_name="input_ids",
    mask_name="input_mask",
    segment_ids_name="segment_ids")

DEFAULT_OUTPUT_NAMES = bert_qa_outputs(
    start_logits_name="start_logits", end_logits_name="end_logits")

_MODEL_INFO = {
    "mobilebert_float.tflite":
        QuestionAnswererInfo(
            name="MobileBert Question and Answerer",
            version="v1",
            description=DEFAULT_DESCRIPTION,
            input_names=DEFAULT_INPUT_NAMES,
            output_names=DEFAULT_OUTPUT_NAMES,
            tokenizer_type=Tokenizer.BERT_TOKENIZER,
            vocab_file="vocab.txt"),
    "albert_float.tflite":
        QuestionAnswererInfo(
            name="Albert Question and Answerer",
            version="v1",
            description=DEFAULT_DESCRIPTION,
            input_names=DEFAULT_INPUT_NAMES,
            output_names=DEFAULT_OUTPUT_NAMES,
            tokenizer_type=Tokenizer.SENTENCE_PIECE,
            vocab_file="30k-clean.vocab",
            sp_model="30k-clean.model"),
}


class MetadataPopulatorForBertQuestionAndAnswer(MetadataPopulatorForBert):
  """Populates the metadata for a Bert QA model."""

  def _create_output_metadata(self):
    """Creates the output metadata for a Bert QA model."""

    # Creates outputs info.
    end_meta = _metadata_fb.TensorMetadataT()
    end_meta.name = "end_logits"
    end_meta.description = (
        "logits over the sequence which indicates the"
        " end position of the answer span with closed interval.")
    end_meta.content = _metadata_fb.ContentT()
    end_meta.content.contentPropertiesType = (
        _metadata_fb.ContentProperties.FeatureProperties)
    end_meta.content.contentProperties = _metadata_fb.FeaturePropertiesT()

    start_meta = _metadata_fb.TensorMetadataT()
    start_meta.name = "start_logits"
    start_meta.description = (
        "logits over the sequence which indicates "
        "the start position of the answer span with closed interval.")
    start_meta.content = _metadata_fb.ContentT()
    start_meta.content.contentPropertiesType = (
        _metadata_fb.ContentProperties.FeatureProperties)
    start_meta.content.contentProperties = _metadata_fb.FeaturePropertiesT()

    # The order of output_metadata should match the order of tensor names in
    # OutputTensorNames.
    output_metadata = [start_meta, end_meta]
    # Order the tensor metadata according to the output tensor order.
    ordered_output_names = self._get_output_tensor_names()
    return self._order_tensor_metadata_with_names(output_metadata,
                                                  self.model_info.output_names,
                                                  ordered_output_names)



# Copyright 2021 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
r"""Run TF Lite model."""
from absl import app
from absl import flags

from PIL import Image
import tensorflow as tf

from tensorflow_examples.lite.model_maker.third_party.efficientdet import inference

FLAGS = flags.FLAGS


def define_flags():
  """Define flags."""
  flags.DEFINE_string('tflite_path', None, 'Path of tflite file.')
  flags.DEFINE_string('sample_image', None, 'Sample image path')
  flags.DEFINE_string('output_image', None, 'Output image path')
  flags.DEFINE_string('image_size', '512x512', 'Image size "WxH".')


def load_image(image_path, image_size):
  """Loads an image, and returns numpy.ndarray.

  Args:
    image_path: str, path to image.
    image_size: list of int, representing [width, height].

  Returns:
    image_batch: numpy.ndarray of shape [1, H, W, C].
  """
  input_data = tf.io.gfile.GFile(image_path, 'rb').read()
  image = tf.io.decode_image(input_data, channels=3, dtype=tf.uint8)
  image = tf.image.resize(
      image, image_size, method='bilinear', antialias=True)
  return tf.expand_dims(tf.cast(image, tf.uint8), 0).numpy()


def save_visualized_image(image, prediction, output_path):
  """Saves the visualized image with prediction.

  Args:
    image: numpy.ndarray of shape [H, W, C].
    prediction: numpy.ndarray of shape [num_predictions, 7].
    output_path: str, output image path.
  """
  output_image = inference.visualize_image_prediction(
      image,
      prediction,
      label_map='coco')
  Image.fromarray(output_image).save(output_path)


class TFLiteRunner:
  """Wrapper to run TFLite model."""

  def __init__(self, model_path):
    """Init.

    Args:
      model_path: str, path to tflite model.
    """
    self.interpreter = tf.lite.Interpreter(model_path=model_path)
    self.interpreter.allocate_tensors()
    self.input_index = self.interpreter.get_input_details()[0]['index']
    self.output_index = self.interpreter.get_output_details()[0]['index']

  def run(self, image):
    """Run inference on a single images.

    Args:
      image: numpy.ndarray of shape [1, H, W, C].

    Returns:
      prediction: numpy.ndarray of shape [1, num_detections, 7].
    """
    self.interpreter.set_tensor(self.input_index, image)
    self.interpreter.invoke()
    return self.interpreter.get_tensor(self.output_index)


def main(_):
  image_size = [int(dim) for dim in FLAGS.image_size.split('x')]
  image = load_image(FLAGS.sample_image, image_size)

  runner = TFLiteRunner(FLAGS.tflite_path)
  prediction = runner.run(image)

  save_visualized_image(image[0], prediction[0], FLAGS.output_image)


if __name__ == '__main__':
  define_flags()
  app.run(main)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A binary that generates saved model artifacts for testing."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import os
import sys

import tensorflow.compat.v2 as tf

def parse_args():
  parser = argparse.ArgumentParser(
      'Generates saved model artifacts for testing.')
  parser.add_argument(
      'output_path',
      type=str,
      help='Model output path.')
  parser.add_argument(
      '--model_type',
      type=str,
      required=True,
      choices=set(['tf_keras_h5', 'tf_saved_model']),
      help='Model format to generate.')
  return parser.parse_known_args()


def main(_):

  if args.model_type == 'tf_keras_h5':
    model = tf.keras.Sequential()
    model.add(tf.keras.layers.Dense(5, activation='relu', input_shape=(8,)))
    model.add(tf.keras.layers.Dense(1, activation='sigmoid'))
    model.save(os.path.join(args.output_path))
  elif args.model_type == 'tf_saved_model':
    class TimesThreePlusOne(tf.Module):

      @tf.function(input_signature=[
          tf.TensorSpec(shape=None, dtype=tf.float32)])
      def compute(self, x):
        return x * 3.0 + 1.0

    tf.saved_model.save(TimesThreePlusOne(), args.output_path)
  else:
    raise ValueError('Unrecognized model type: %s' % args.model_type)


if __name__ == '__main__':
  args, unparsed = parse_args()
  tf.compat.v1.app.run(main=main, argv=[sys.argv[0]] + unparsed)



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Recommendation demo code of Model Maker for TFLite."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from absl import app
from absl import flags
from absl import logging

from tflite_model_maker import model_spec as ms
from tflite_model_maker import recommendation

FLAGS = flags.FLAGS


def define_flags():
  flags.DEFINE_string('data_dir', None, 'The directory to save dataset.')
  flags.DEFINE_string('export_dir', None,
                      'The directory to save exported files.')
  flags.DEFINE_string('encoder_type', 'bow',
                      'The recommendation encoder to run. (bow, cnn, lstm)')
  flags.mark_flag_as_required('export_dir')


def download_data(download_dir):
  """Downloads demo data, and returns directory path."""
  return recommendation.DataLoader.download_and_extract_movielens(download_dir)


def get_input_spec(encoder_type: str,
                   num_classes: int) -> recommendation.spec.InputSpec:
  """Gets input spec (for test).

  Input spec defines how the input features are extracted.

  Args:
    encoder_type: str, case-insensitive {'CNN', 'LSTM', 'BOW'}.
    num_classes: int, num of classes in vocabulary.

  Returns:
    InputSpec.
  """
  etype = encoder_type.upper()
  if etype not in {'CNN', 'LSTM', 'BOW'}:
    raise ValueError('Not support encoder_type: {}'.format(etype))

  return recommendation.spec.InputSpec(
      activity_feature_groups=[
          # Group #1: defines how features are grouped in the first Group.
          dict(
              features=[
                  # First feature.
                  dict(
                      feature_name='context_movie_id',  # Feature name
                      feature_type='INT',  # Feature type
                      vocab_size=num_classes,  # ID size (number of IDs)
                      embedding_dim=8,  # Projected feature embedding dim
                      feature_length=10,  # History length of 10.
                  ),
                  # Maybe more features...
              ],
              encoder_type='CNN',  # CNN encoder (e.g. CNN, LSTM, BOW)
          ),
          # Maybe more groups...
      ],
      label_feature=dict(
          feature_name='label_movie_id',  # Label feature name
          feature_type='INT',  # Label type
          vocab_size=num_classes,  # Label size (number of classes)
          embedding_dim=8,  # Label embedding demension
          feature_length=1,  # Exactly 1 label
      ),
  )


def get_model_hparams() -> recommendation.spec.ModelHParams:
  """Gets model hparams (for test).

  ModelHParams defines the model architecture.

  Returns:
    ModelHParams.
  """
  return recommendation.spec.ModelHParams(
      hidden_layer_dims=[32, 32],  # Hidden layers dimension.
      eval_top_k=[1, 5],  # Eval top 1 and top 5.
      conv_num_filter_ratios=[2, 4],  # For CNN encoder, conv filter mutipler.
      conv_kernel_size=16,  # For CNN encoder, base kernel size.
      lstm_num_units=16,  # For LSTM/RNN, num units.
      num_predictions=10,  # Number of output predictions. Select top 10.
  )


def run(data_dir, export_dir, batch_size=16, epochs=5, encoder_type='bow'):
  """Runs demo."""
  meta = recommendation.DataLoader.generate_movielens_dataset(data_dir)
  num_classes = recommendation.DataLoader.get_num_classes(meta)

  input_spec = get_input_spec(encoder_type, num_classes)
  train_data = recommendation.DataLoader.from_movielens(data_dir, 'train',
                                                        input_spec)
  test_data = recommendation.DataLoader.from_movielens(data_dir, 'test',
                                                       input_spec)

  model_spec = ms.get(
      'recommendation',
      input_spec=input_spec,
      model_hparams=get_model_hparams())
  # Create a model and train.
  model = recommendation.create(
      train_data,
      model_spec=model_spec,
      model_dir=export_dir,
      validation_data=test_data,
      batch_size=batch_size,
      epochs=epochs)

  # Evaluate with test_data.
  history = model.evaluate(test_data)
  print('Test metrics from Keras model: %s' % history)

  # Export tflite model.
  model.export(export_dir)

  # Evaluate tflite model.
  tflite_model = os.path.join(export_dir, 'model.tflite')
  history = model.evaluate_tflite(tflite_model, test_data)
  print('Test metrics from TFLite model: %s' % history)


def main(_):
  logging.set_verbosity(logging.INFO)

  export_dir = os.path.expanduser(FLAGS.export_dir)
  data_dir = os.path.expanduser(FLAGS.data_dir)

  extracted_dir = download_data(data_dir)
  run(extracted_dir, export_dir, encoder_type=FLAGS.encoder_type)


if __name__ == '__main__':
  define_flags()
  app.run(main)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Artifact conversion to and from Python TensorFlow and tf.keras."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import json
import os
import shutil
import sys
import tempfile

import h5py
import tensorflow.compat.v1 as tf1
import tensorflow.compat.v2 as tf

from tensorflowjs import quantization
from tensorflowjs import version
from tensorflowjs.converters import common
from tensorflowjs.converters import keras_h5_conversion as conversion
from tensorflowjs.converters import keras_tfjs_loader
from tensorflowjs.converters import tf_saved_model_conversion_v2


def dispatch_keras_h5_to_tfjs_layers_model_conversion(
    h5_path, output_dir=None, quantization_dtype_map=None,
    split_weights_by_layer=False,
    weight_shard_size_bytes=1024 * 1024 * 4,
    metadata=None):
  """Converts a Keras HDF5 saved-model file to TensorFlow.js format.

  Auto-detects saved_model versus weights-only and generates the correct
  json in either case. This function accepts Keras HDF5 files in two formats:
    - A weights-only HDF5 (e.g., generated with Keras Model's `save_weights()`
      method),
    - A topology+weights combined HDF5 (e.g., generated with
      `tf.keras.model.save_model`).

  Args:
    h5_path: path to an HDF5 file containing keras model data as a `str`.
    output_dir: Output directory to which the TensorFlow.js-format model JSON
      file and weights files will be written. If the directory does not exist,
      it will be created.
    quantization_dtype_map: A mapping from dtype (`uint8`, `uint16`, `float16`)
      to weights. The weight mapping supports wildcard substitution.
    split_weights_by_layer: Whether to split the weights into separate weight
      groups (corresponding to separate binary weight files) layer by layer
      (Default: `False`).
    weight_shard_size_bytes: Shard size (in bytes) of the weight files.
      The size of each weight file will be <= this value.
    metadata: User defined metadata map.

  Returns:
    (model_json, groups)
      model_json: a json dictionary (empty if unused) for model topology.
        If `h5_path` points to a weights-only HDF5 file, this return value
        will be `None`.
      groups: an array of weight_groups as defined in tfjs weights_writer.
  """
  if not os.path.exists(h5_path):
    raise ValueError('Nonexistent path to HDF5 file: %s' % h5_path)
  if os.path.isdir(h5_path):
    raise ValueError(
        'Expected path to point to an HDF5 file, but it points to a '
        'directory: %s' % h5_path)

  h5_file = h5py.File(h5_path, 'r')
  if 'layer_names' in h5_file.attrs:
    model_json = None
    groups = conversion.h5_weights_to_tfjs_format(
        h5_file, split_by_layer=split_weights_by_layer)
  else:
    model_json, groups = conversion.h5_merged_saved_model_to_tfjs_format(
        h5_file, split_by_layer=split_weights_by_layer)

  if output_dir:
    if os.path.isfile(output_dir):
      raise ValueError(
          'Output path "%s" already exists as a file' % output_dir)
    if not os.path.isdir(output_dir):
      os.makedirs(output_dir)
    conversion.write_artifacts(
        model_json, groups, output_dir, quantization_dtype_map,
        weight_shard_size_bytes=weight_shard_size_bytes, metadata=metadata)

  return model_json, groups


def dispatch_keras_h5_to_tfjs_graph_model_conversion(
    h5_path, output_dir=None,
    quantization_dtype_map=None,
    skip_op_check=False,
    strip_debug_ops=False,
    weight_shard_size_bytes=1024 * 1024 * 4,
    control_flow_v2=False,
    experiments=False,
    metadata=None):
  """
  Convert a keras HDF5-format model to tfjs GraphModel artifacts.

  Args:
    h5_path: Path to the HDF5-format file that contains the model saved from
      keras or tf.keras.
    output_dir: The destination to which the tfjs GraphModel artifacts will be
      written.
    quantization_dtype_map: A mapping from dtype (`uint8`, `uint16`, `float16`)
      to weights. The weight mapping supports wildcard substitution.
    skip_op_check: Bool whether to skip the op check.
    strip_debug_ops: Bool whether to allow unsupported debug ops.
    weight_shard_size_bytes: Shard size (in bytes) of the weight files.
      The size of each weight file will be <= this value.
    control_flow_v2: Bool whether to enable control flow v2 ops.
    experiments: Bool enable experimental features.
    metadata: User defined metadata map.
  """

  if not os.path.exists(h5_path):
    raise ValueError('Nonexistent path to HDF5 file: %s' % h5_path)
  if os.path.isdir(h5_path):
    raise ValueError(
        'Expected path to point to an HDF5 file, but it points to a '
        'directory: %s' % h5_path)

  temp_savedmodel_dir = tempfile.mktemp(suffix='.savedmodel')
  model = tf.keras.models.load_model(h5_path, compile=False)
  model.save(temp_savedmodel_dir, include_optimizer=False, save_format='tf')

  # NOTE(cais): This cannot use `tf.compat.v1` because
  #   `convert_tf_saved_model()` works only in v2.
  tf_saved_model_conversion_v2.convert_tf_saved_model(
      temp_savedmodel_dir, output_dir,
      signature_def='serving_default',
      saved_model_tags='serve',
      quantization_dtype_map=quantization_dtype_map,
      skip_op_check=skip_op_check,
      strip_debug_ops=strip_debug_ops,
      weight_shard_size_bytes=weight_shard_size_bytes,
      control_flow_v2=control_flow_v2,
      experiments=experiments,
      metadata=metadata)

  # Clean up the temporary SavedModel directory.
  shutil.rmtree(temp_savedmodel_dir)


def dispatch_keras_saved_model_to_tensorflowjs_conversion(
    keras_saved_model_path, output_dir, quantization_dtype_map=None,
    split_weights_by_layer=False,
    weight_shard_size_bytes=1024 * 1024 * 4,
    metadata=None):
  """Converts keras model saved in the SavedModel format to tfjs format.

  Note that the SavedModel format exists in keras, but not in
  keras-team/tf.keras.

  Args:
    keras_saved_model_path: path to a folder in which the
      assets/saved_model.json can be found. This is usually a subfolder
      that is under the folder passed to
      `tf.keras.models.save_model()` and has a Unix epoch time
      as its name (e.g., 1542212752).
    output_dir: Output directory to which the TensorFlow.js-format model JSON
      file and weights files will be written. If the directory does not exist,
      it will be created.
    quantization_dtype_map: A mapping from dtype (`uint8`, `uint16`, `float16`)
      to weights. The weight mapping supports wildcard substitution.
    split_weights_by_layer: Whether to split the weights into separate weight
      groups (corresponding to separate binary weight files) layer by layer
      (Default: `False`).
    weight_shard_size_bytes: Shard size (in bytes) of the weight files.
      The size of each weight file will be <= this value.
    metadata: User defined metadata map.
  """
  with tf.Graph().as_default(), tf.compat.v1.Session():
    model = tf.keras.models.load_model(keras_saved_model_path)

    # Save model temporarily in HDF5 format.
    temp_h5_path = tempfile.mktemp(suffix='.h5')
    model.save(temp_h5_path, save_format='h5')
    assert os.path.isfile(temp_h5_path)

    dispatch_keras_h5_to_tfjs_layers_model_conversion(
        temp_h5_path,
        output_dir,
        quantization_dtype_map=quantization_dtype_map,
        split_weights_by_layer=split_weights_by_layer,
        weight_shard_size_bytes=weight_shard_size_bytes,
        metadata=metadata)

    # Delete temporary .h5 file.
    os.remove(temp_h5_path)


def dispatch_tensorflowjs_to_keras_h5_conversion(config_json_path, h5_path):
  """Converts a TensorFlow.js Layers model format to Keras H5.

  Args:
    config_json_path: Path to the JSON file that includes the model's
      topology and weights manifest, in tensorflowjs format.
    h5_path: Path for the to-be-created Keras HDF5 model file.

  Raises:
    ValueError, if `config_json_path` is not a path to a valid JSON
      file, or if h5_path points to an existing directory.
  """
  if os.path.isdir(config_json_path):
    raise ValueError(
        'For input_type=tfjs_layers_model & output_format=keras, '
        'the input path should be a model.json '
        'file, but received a directory.')
  if os.path.isdir(h5_path):
    raise ValueError(
        'For input_type=tfjs_layers_model & output_format=keras, '
        'the output path should be the path to an HDF5 file, '
        'but received an existing directory (%s).' % h5_path)

  # Verify that config_json_path points to a JSON file.
  with open(config_json_path, 'rt') as f:
    try:
      json.load(f)
    except (ValueError, IOError):
      raise ValueError(
          'For input_type=tfjs_layers_model & output_format=keras, '
          'the input path is expected to contain valid JSON content, '
          'but cannot read valid JSON content from %s.' % config_json_path)

  with tf.Graph().as_default(), tf.compat.v1.Session():
    model = keras_tfjs_loader.load_keras_model(config_json_path)
    model.save(h5_path)


def dispatch_tensorflowjs_to_keras_saved_model_conversion(
    config_json_path, keras_saved_model_path):
  """Converts a TensorFlow.js Layers model format to a tf.keras SavedModel.

  Args:
    config_json_path: Path to the JSON file that includes the model's
      topology and weights manifest, in tensorflowjs format.
    keras_saved_model_path: Path for the to-be-created Keras SavedModel.

  Raises:
    ValueError, if `config_json_path` is not a path to a valid JSON
      file, or if h5_path points to an existing directory.
  """
  if os.path.isdir(config_json_path):
    raise ValueError(
        'For input_type=tfjs_layers_model & output_format=keras_saved_model, '
        'the input path should be a model.json '
        'file, but received a directory.')

  # Verify that config_json_path points to a JSON file.
  with open(config_json_path, 'rt') as f:
    try:
      json.load(f)
    except (ValueError, IOError):
      raise ValueError(
          'For input_type=tfjs_layers_model & output_format=keras, '
          'the input path is expected to contain valid JSON content, '
          'but cannot read valid JSON content from %s.' % config_json_path)

  with tf.Graph().as_default(), tf.compat.v1.Session():
    model = keras_tfjs_loader.load_keras_model(config_json_path)
    tf.keras.models.save_model(
        model, keras_saved_model_path, save_format='tf')


def dispatch_tensorflowjs_to_tensorflowjs_conversion(
    config_json_path,
    output_dir_path,
    quantization_dtype_map=None,
    weight_shard_size_bytes=1024 * 1024 * 4):
  """Converts a Keras Model from tensorflowjs format to H5.

  Args:
    config_json_path: Path to the JSON file that includes the model's
      topology and weights manifest, in tensorflowjs format.
    output_dir_path: Path to output directory in which the result of the
      conversion will be saved.
    quantization_dtype_map: A mapping from dtype (`uint8`, `uint16`, `float16`)
      to weights. The weight mapping supports wildcard substitution.
    weight_shard_size_bytes: Shard size (in bytes) of the weight files.
      The size of each weight file will be <= this value.

  Raises:
    ValueError, if `config_json_path` is not a path to a valid JSON
      file, or if h5_path points to an existing directory.
    ValueError, if `output_dir_path` exists and is a file (instead of
      a directory).
  """
  if os.path.isdir(config_json_path):
    raise ValueError(
        'For input_type=tfjs_layers_model, '
        'the input path should be a model.json '
        'file, but received a directory.')
  # TODO(cais): Assert output_dir_path doesn't exist or is the path to
  # a directory (not a file).

  # Verify that config_json_path points to a JSON file.
  with open(config_json_path, 'rt') as f:
    try:
      json.load(f)
    except (ValueError, IOError):
      raise ValueError(
          'For input_type=tfjs_layers_model, '
          'the input path is expected to contain valid JSON content, '
          'but cannot read valid JSON content from %s.' % config_json_path)

  temp_h5_path = tempfile.mktemp(suffix='.h5')
  with tf.Graph().as_default(), tf.compat.v1.Session():
    model = keras_tfjs_loader.load_keras_model(config_json_path)
    model.save(temp_h5_path)
    dispatch_tensorflowjs_to_keras_h5_conversion(config_json_path, temp_h5_path)

  with tf.Graph().as_default(), tf.compat.v1.Session():
    dispatch_keras_h5_to_tfjs_layers_model_conversion(
        temp_h5_path, output_dir_path,
        quantization_dtype_map=quantization_dtype_map,
        weight_shard_size_bytes=weight_shard_size_bytes)
    # TODO(cais): Support weight quantization.

  # Clean up the temporary H5 file.
  os.remove(temp_h5_path)


def dispatch_tfjs_layers_model_to_tfjs_graph_conversion(
    config_json_path,
    output_dir_path,
    quantization_dtype_map=None,
    skip_op_check=False,
    strip_debug_ops=False,
    weight_shard_size_bytes=1024 * 1024 * 4,
    control_flow_v2=False,
    experiments=False,
    metadata=None):
  """Converts a TensorFlow.js Layers Model to TensorFlow.js Graph Model.

  This conversion often benefits speed of inference, due to the graph
  optimization that goes into generating the Graph Model.

  Args:
    config_json_path: Path to the JSON file that includes the model's
      topology and weights manifest, in tensorflowjs format.
    output_dir_path: Path to output directory in which the result of the
      conversion will be saved.
    quantization_dtype_map: A mapping from dtype (`uint8`, `uint16`, `float16`)
      to weights. The weight mapping supports wildcard substitution.
    skip_op_check: Bool whether to skip the op check.
    strip_debug_ops: Bool whether to allow unsupported debug ops.
    weight_shard_size_bytes: Shard size (in bytes) of the weight files.
      The size of each weight file will be <= this value.
    control_flow_v2: Bool whether to enable control flow v2 ops.
    experiments: Bool enable experimental features.
    metadata: User defined metadata map.
  Raises:
    ValueError, if `config_json_path` is not a path to a valid JSON
      file, or if h5_path points to an existing directory.
    ValueError, if `output_dir_path` exists and is a file (instead of
      a directory).
  """
  if os.path.isdir(config_json_path):
    raise ValueError(
        'For input_type=tfjs_layers_model, '
        'the input path should be a model.json '
        'file, but received a directory.')
  # TODO(cais): Assert output_dir_path doesn't exist or is the path to
  # a directory (not a file).

  # Verify that config_json_path points to a JSON file.
  with open(config_json_path, 'rt') as f:
    try:
      json.load(f)
    except (ValueError, IOError):
      raise ValueError(
          'For input_type=tfjs_layers_model, '
          'the input path is expected to contain valid JSON content, '
          'but cannot read valid JSON content from %s.' % config_json_path)

  temp_h5_path = tempfile.mktemp(suffix='.h5')

  model = keras_tfjs_loader.load_keras_model(config_json_path)
  model.save(temp_h5_path)
  dispatch_keras_h5_to_tfjs_graph_model_conversion(
      temp_h5_path, output_dir_path,
      quantization_dtype_map=quantization_dtype_map,
      skip_op_check=skip_op_check,
      strip_debug_ops=strip_debug_ops,
      weight_shard_size_bytes=weight_shard_size_bytes,
      control_flow_v2=control_flow_v2,
      experiments=experiments,
      metadata=metadata)

  # Clean up temporary HDF5 file.
  os.remove(temp_h5_path)


def _standardize_input_output_formats(input_format, output_format):
  """Standardize input and output formats.

  Args:
    input_format: Input format as a string.
    output_format: Output format as a string.

  Returns:
    A `tuple` of two strings:
      (standardized_input_format, standardized_output_format).
  """
  input_format_is_keras = (
      input_format in [common.KERAS_MODEL, common.KERAS_SAVED_MODEL])
  input_format_is_tf = (
      input_format in [common.TF_SAVED_MODEL,
                       common.TF_FROZEN_MODEL, common.TF_HUB_MODEL])
  if output_format is None:
    # If no explicit output_format is provided, infer it from input format.
    if input_format_is_keras:
      output_format = common.TFJS_LAYERS_MODEL
    elif input_format_is_tf:
      output_format = common.TFJS_GRAPH_MODEL
    elif input_format == common.TFJS_LAYERS_MODEL:
      output_format = common.KERAS_MODEL

  return (input_format, output_format)

def _parse_quantization_dtype_map(float16, uint8, uint16, quantization_bytes):
  quantization_dtype_map = {}

  if quantization_bytes:
    print(
        'Warning: --quantization_bytes will be deprecated in a future release\n'
        'Please consider using --quantize_uint8, --quantize_uint16, '
        '--quantize_float16.', file=sys.stderr)
    if float16 is not None or uint8 is not None or uint16 is not None:
      raise ValueError(
          '--quantization_bytes cannot be used with the new quantization flags')

    dtype = quantization.QUANTIZATION_BYTES_TO_DTYPES[quantization_bytes]
    quantization_dtype_map[dtype] = True

  if float16 is not None:
    quantization_dtype_map[quantization.QUANTIZATION_DTYPE_FLOAT16] = \
      float16.split(',') if isinstance(float16, str) else float16
  if uint8 is not None:
    quantization_dtype_map[quantization.QUANTIZATION_DTYPE_UINT8] = \
      uint8.split(',') if isinstance(uint8, str) else uint8
  if uint16 is not None:
    quantization_dtype_map[quantization.QUANTIZATION_DTYPE_UINT16] = \
      uint16.split(',') if isinstance(uint16, str) else uint16

  return quantization_dtype_map

def _parse_metadata_map(metadata_arg):
  if metadata_arg:
    metadata_map = {}
    metadata_list = metadata_arg.split(',')
    for metadata in metadata_list:
      [key, path] = metadata.split(':')
      with open(path, 'rt') as f:
        try:
          metadata_json = json.load(f)
          metadata_map[key] = metadata_json
        except (ValueError, IOError):
          raise ValueError(
              'Loading metadata error: '
              'the path is expected to contain valid JSON content, '
              'but cannot read valid JSON content from %s.' % path)
    if not metadata_map:
      metadata_map = None
    return metadata_map

  return None

def _dispatch_converter(input_format,
                        output_format,
                        args,
                        quantization_dtype_map,
                        weight_shard_size_bytes,
                        metadata_map):
  # TODO(cais, piyu): More conversion logics can be added as additional
  #   branches below.
  if (input_format == common.KERAS_MODEL and
      output_format == common.TFJS_LAYERS_MODEL):
    dispatch_keras_h5_to_tfjs_layers_model_conversion(
        args.input_path, output_dir=args.output_path,
        quantization_dtype_map=quantization_dtype_map,
        split_weights_by_layer=args.split_weights_by_layer,
        weight_shard_size_bytes=weight_shard_size_bytes,
        metadata=metadata_map)
  elif (input_format == common.KERAS_MODEL and
        output_format == common.TFJS_GRAPH_MODEL):
    dispatch_keras_h5_to_tfjs_graph_model_conversion(
        args.input_path, output_dir=args.output_path,
        quantization_dtype_map=quantization_dtype_map,
        skip_op_check=args.skip_op_check,
        strip_debug_ops=args.strip_debug_ops,
        weight_shard_size_bytes=weight_shard_size_bytes,
        control_flow_v2=args.control_flow_v2,
        experiments=args.experiments,
        metadata=metadata_map)
  elif (input_format == common.KERAS_SAVED_MODEL and
        output_format == common.TFJS_LAYERS_MODEL):
    dispatch_keras_saved_model_to_tensorflowjs_conversion(
        args.input_path, args.output_path,
        quantization_dtype_map=quantization_dtype_map,
        split_weights_by_layer=args.split_weights_by_layer,
        weight_shard_size_bytes=weight_shard_size_bytes,
        metadata=metadata_map)
  elif (input_format == common.TF_SAVED_MODEL and
        output_format == common.TFJS_GRAPH_MODEL):
    tf_saved_model_conversion_v2.convert_tf_saved_model(
        args.input_path, args.output_path,
        signature_def=args.signature_name,
        saved_model_tags=args.saved_model_tags,
        quantization_dtype_map=quantization_dtype_map,
        skip_op_check=args.skip_op_check,
        strip_debug_ops=args.strip_debug_ops,
        weight_shard_size_bytes=weight_shard_size_bytes,
        control_flow_v2=args.control_flow_v2,
        experiments=args.experiments,
        metadata=metadata_map)
  elif (input_format == common.TF_HUB_MODEL and
        output_format == common.TFJS_GRAPH_MODEL):
    tf_saved_model_conversion_v2.convert_tf_hub_module(
        args.input_path, args.output_path,
        signature=args.signature_name,
        saved_model_tags=args.saved_model_tags,
        quantization_dtype_map=quantization_dtype_map,
        skip_op_check=args.skip_op_check,
        strip_debug_ops=args.strip_debug_ops,
        weight_shard_size_bytes=weight_shard_size_bytes,
        control_flow_v2=args.control_flow_v2,
        experiments=args.experiments,
        metadata=metadata_map)
  elif (input_format == common.TFJS_LAYERS_MODEL and
        output_format == common.KERAS_MODEL):
    dispatch_tensorflowjs_to_keras_h5_conversion(args.input_path,
                                                 args.output_path)
  elif (input_format == common.TFJS_LAYERS_MODEL and
        output_format == common.KERAS_SAVED_MODEL):
    dispatch_tensorflowjs_to_keras_saved_model_conversion(args.input_path,
                                                          args.output_path)
  elif (input_format == common.TFJS_LAYERS_MODEL and
        output_format == common.TFJS_LAYERS_MODEL):
    dispatch_tensorflowjs_to_tensorflowjs_conversion(
        args.input_path, args.output_path,
        quantization_dtype_map=quantization_dtype_map,
        weight_shard_size_bytes=weight_shard_size_bytes)
  elif (input_format == common.TFJS_LAYERS_MODEL and
        output_format == common.TFJS_GRAPH_MODEL):
    dispatch_tfjs_layers_model_to_tfjs_graph_conversion(
        args.input_path, args.output_path,
        quantization_dtype_map=quantization_dtype_map,
        skip_op_check=args.skip_op_check,
        strip_debug_ops=args.strip_debug_ops,
        weight_shard_size_bytes=weight_shard_size_bytes,
        control_flow_v2=args.control_flow_v2,
        experiments=args.experiments,
        metadata=metadata_map)
  elif (input_format == common.TF_FROZEN_MODEL and
        output_format == common.TFJS_GRAPH_MODEL):
    tf_saved_model_conversion_v2.convert_tf_frozen_model(
        args.input_path, args.output_node_names, args.output_path,
        quantization_dtype_map=quantization_dtype_map,
        skip_op_check=args.skip_op_check,
        strip_debug_ops=args.strip_debug_ops,
        weight_shard_size_bytes=weight_shard_size_bytes,
        experiments=args.experiments,
        metadata=metadata_map)
  else:
    raise ValueError(
        'Unsupported input_format - output_format pair: %s - %s' %
        (input_format, output_format))

def get_arg_parser():
  """
  Create the argument parser for the converter binary.
  """

  parser = argparse.ArgumentParser('TensorFlow.js model converters.')
  parser.add_argument(
      common.INPUT_PATH,
      nargs='?',
      type=str,
      help='Path to the input file or directory. For input format "keras", '
      'an HDF5 (.h5) file is expected. For input format "tensorflow", '
      'a SavedModel directory, frozen model file, '
      'or TF-Hub module is expected.')
  parser.add_argument(
      common.OUTPUT_PATH,
      nargs='?',
      type=str,
      help='Path for all output artifacts.')
  parser.add_argument(
      '--%s' % common.INPUT_FORMAT,
      type=str,
      required=False,
      default=common.TF_SAVED_MODEL,
      choices=set([common.KERAS_MODEL, common.KERAS_SAVED_MODEL,
                   common.TF_SAVED_MODEL, common.TF_HUB_MODEL,
                   common.TFJS_LAYERS_MODEL, common.TF_FROZEN_MODEL]),
      help='Input format. '
      'For "keras", the input path can be one of the two following formats:\n'
      '  - A topology+weights combined HDF5 (e.g., generated with'
      '    `tf.keras.model.save_model()` method).\n'
      '  - A weights-only HDF5 (e.g., generated with Keras Model\'s '
      '    `save_weights()` method). \n'
      'For "keras_saved_model", the input_path must point to a subfolder '
      'under the saved model folder that is passed as the argument '
      'to tf.contrib.save_model.save_keras_model(). '
      'The subfolder is generated automatically by tensorflow when '
      'saving keras model in the SavedModel format. It is usually named '
      'as a Unix epoch time (e.g., 1542212752).\n'
      'For "tf" formats, a SavedModel, frozen model, '
      ' or TF-Hub module is expected.')
  parser.add_argument(
      '--%s' % common.OUTPUT_FORMAT,
      type=str,
      required=False,
      choices=set([common.KERAS_MODEL, common.KERAS_SAVED_MODEL,
                   common.TFJS_LAYERS_MODEL, common.TFJS_GRAPH_MODEL]),
      help='Output format. Default: tfjs_graph_model.')
  parser.add_argument(
      '--%s' % common.SIGNATURE_NAME,
      type=str,
      default=None,
      help='Signature of the SavedModel Graph or TF-Hub module to load. '
      'Applicable only if input format is "tf_hub" or "tf_saved_model".')
  parser.add_argument(
      '--%s' % common.SAVED_MODEL_TAGS,
      type=str,
      default='serve',
      help='Tags of the MetaGraphDef to load, in comma separated string '
      'format. Defaults to "serve". Applicable only if input format is '
      '"tf_saved_model".')
  parser.add_argument(
      '--%s' % common.QUANTIZATION_TYPE_FLOAT16,
      type=str,
      default=None,
      const=True,
      nargs='?',
      help='Comma separated list of node names to apply float16 quantization. '
      'You can also use wildcard symbol (*) to apply quantization to multiple '
      'nodes (e.g., conv/*/weights). When the flag is provided without any '
      'nodes the default behavior will match all nodes.')
  parser.add_argument(
      '--%s' % common.QUANTIZATION_TYPE_UINT8,
      type=str,
      default=None,
      const=True,
      nargs='?',
      help='Comma separated list of node names to apply 1-byte affine '
      'quantization. You can also use wildcard symbol (*) to apply '
      'quantization to multiple nodes (e.g., conv/*/weights). When the flag is '
      'provided without any nodes the default behavior will match all nodes.')
  parser.add_argument(
      '--%s' % common.QUANTIZATION_TYPE_UINT16,
      type=str,
      default=None,
      const=True,
      nargs='?',
      help='Comma separated list of node names to apply 2-byte affine '
      'quantization. You can also use wildcard symbol (*) to apply '
      'quantization to multiple nodes (e.g., conv/*/weights). When the flag is '
      'provided without any nodes the default behavior will match all nodes.')
  parser.add_argument(
      '--%s' % common.QUANTIZATION_BYTES,
      type=int,
      choices=set(quantization.QUANTIZATION_BYTES_TO_DTYPES.keys()),
      help='(Deprecated) How many bytes to optionally quantize/compress the '
      'weights to. 1- and 2-byte quantizaton is supported. The default '
      '(unquantized) size is 4 bytes.')
  parser.add_argument(
      '--%s' % common.SPLIT_WEIGHTS_BY_LAYER,
      action='store_true',
      help='Applicable to keras input_format only: Whether the weights from '
      'different layers are to be stored in separate weight groups, '
      'corresponding to separate binary weight files. Default: False.')
  parser.add_argument(
      '--%s' % common.VERSION,
      '-v',
      dest='show_version',
      action='store_true',
      help='Show versions of tensorflowjs and its dependencies')
  parser.add_argument(
      '--%s' % common.SKIP_OP_CHECK,
      action='store_true',
      help='Skip op validation for TensorFlow model conversion.')
  parser.add_argument(
      '--%s' % common.STRIP_DEBUG_OPS,
      type=bool,
      default=True,
      help='Strip debug ops (Print, Assert, CheckNumerics) from graph.')
  parser.add_argument(
      '--%s' % common.WEIGHT_SHARD_SIZE_BYTES,
      type=int,
      default=None,
      help='Shard size (in bytes) of the weight files. Currently applicable '
      'only when output_format is tfjs_layers_model or tfjs_graph_model.')
  parser.add_argument(
      '--output_node_names',
      type=str,
      help='The names of the output nodes, separated by commas. E.g., '
      '"logits,activations". Applicable only if input format is '
      '"tf_frozen_model".')
  parser.add_argument(
      '--%s' % common.CONTROL_FLOW_V2,
      type=bool,
      default=False,
      help='Enable control flow v2 ops, this would improve inference '
      'performance on models with branches or loops.')
  parser.add_argument(
      '--%s' % common.EXPERIMENTS,
      type=bool,
      default=False,
      help='Enable experimental features, you should only enable this flag '
      'when using Python3 and TensorFlow nightly build.')
  parser.add_argument(
      '--%s' % common.METADATA,
      type=str,
      help='Attach user defined metadata in format key:path/metadata.json '
      'Separate multiple metadata files by comma.'
  )
  return parser

def convert(arguments):
  args = get_arg_parser().parse_args(arguments)
  if args.show_version:
    print('\ntensorflowjs %s\n' % version.version)
    print('Dependency versions:')
    print('  keras %s' % tf.keras.__version__)
    print('  tensorflow %s' % tf.__version__)
    return

  if not args.input_path:
    raise ValueError(
        'Missing input_path argument. For usage, use the --help flag.')
  if not args.output_path:
    raise ValueError(
        'Missing output_path argument. For usage, use the --help flag.')

  if args.input_path is None:
    raise ValueError(
        'Error: The input_path argument must be set. '
        'Run with --help flag for usage information.')

  input_format, output_format = _standardize_input_output_formats(
      args.input_format, args.output_format)

  weight_shard_size_bytes = 1024 * 1024 * 4
  if args.weight_shard_size_bytes is not None:
    if (output_format not in
        (common.TFJS_LAYERS_MODEL, common.TFJS_GRAPH_MODEL)):
      raise ValueError(
          'The --weight_shard_size_bytes flag is only supported when '
          'output_format is tfjs_layers_model or tfjs_graph_model.')

    if not (isinstance(args.weight_shard_size_bytes, int) and
            args.weight_shard_size_bytes > 0):
      raise ValueError(
          'Expected weight_shard_size_bytes to be a positive integer, '
          'but got %s' % args.weight_shard_size_bytes)
    weight_shard_size_bytes = args.weight_shard_size_bytes

  quantization_dtype_map = _parse_quantization_dtype_map(
      args.quantize_float16,
      args.quantize_uint8,
      args.quantize_uint16,
      args.quantization_bytes
  )

  if (not args.output_node_names and input_format == common.TF_FROZEN_MODEL):
    raise ValueError(
        'The --output_node_names flag is required for "tf_frozen_model"')

  if (args.signature_name and input_format not in
      (common.TF_SAVED_MODEL, common.TF_HUB_MODEL)):
    raise ValueError(
        'The --signature_name flag is applicable only to "tf_saved_model" and '
        '"tf_hub" input format, but the current input format is '
        '"%s".' % input_format)

  if (args.control_flow_v2 and output_format != common.TFJS_GRAPH_MODEL):
    raise ValueError(
        'The --control_flow_v2 flag is applicable only to "tfjs_graph_model" '
        'as output format, but the current  output format '
        'is "%s"' % input_format, output_format)

  metadata_map = _parse_metadata_map(args.metadata)

  _dispatch_converter(input_format, output_format, args, quantization_dtype_map,
                      weight_shard_size_bytes, metadata_map)

def pip_main():
  """Entry point for pip-packaged binary.

  Note that pip-packaged binary calls the entry method without
  any arguments, which is why this method is needed in addition to the
  `main` method below.
  """
  main([' '.join(sys.argv[1:])])


def main(argv):
  convert(argv[0].split(' '))


if __name__ == '__main__':
  tf1.app.run(main=main, argv=[' '.join(sys.argv[1:])])



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Text classification demo code of model customization for TFLite."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from absl import app
from absl import flags
from absl import logging

import tensorflow.compat.v2 as tf
from tflite_model_maker import model_spec
from tflite_model_maker import question_answer

FLAGS = flags.FLAGS


def define_flags():
  flags.DEFINE_string('export_dir', None,
                      'The directory to save exported files.')
  flags.mark_flag_as_required('export_dir')
  flags.DEFINE_string('spec', 'bert_qa', 'The QA model to run.')


def download_demo_data(**kwargs):
  """Downloads demo data, and returns directory path."""
  train_data_path = tf.keras.utils.get_file(
      fname='train-v1.1.json',
      origin='https://rajpurkar.github.io/SQuAD-explorer/dataset/train-v1.1.json',
      **kwargs)
  validation_data_path = tf.keras.utils.get_file(
      fname='dev-v1.1.json',
      origin='https://rajpurkar.github.io/SQuAD-explorer/dataset/dev-v1.1.json',
      **kwargs)
  return train_data_path, validation_data_path


def run(train_data_path,
        validation_data_path,
        export_dir,
        spec='bert_qa',
        **kwargs):
  """Runs demo."""
  # Chooses model specification that represents model.
  spec = model_spec.get(spec)

  # Gets training data and validation data.
  train_data = question_answer.DataLoader.from_squad(
      train_data_path, spec, is_training=True)
  validation_data = question_answer.DataLoader.from_squad(
      validation_data_path, spec, is_training=False)

  # Fine-tunes the model.
  model = question_answer.create(train_data, model_spec=spec, **kwargs)

  # Gets evaluation results.
  metric = model.evaluate(validation_data)
  tf.compat.v1.logging.info('Eval F1 score:%f' % metric['final_f1'])

  # Exports to TFLite format.
  model.export(export_dir)


def main(_):
  logging.set_verbosity(logging.INFO)

  train_data_path, validation_data_path = download_demo_data()
  export_dir = os.path.expanduser(FLAGS.export_dir)

  run(train_data_path, validation_data_path, export_dir, spec=FLAGS.spec)


if __name__ == '__main__':
  define_flags()
  app.run(main)



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=g-bad-import-order,redefined-builtin
"""APIs to train a model that can answer questions based on a predefined text.

Task guide:
https://www.tensorflow.org/lite/tutorials/model_maker_question_answer.
"""

from tensorflow_examples.lite.model_maker.core.data_util.text_dataloader import QuestionAnswerDataLoader as DataLoader
from tensorflow_examples.lite.model_maker.core.task.model_spec.text_spec import BertQAModelSpec as BertQaSpec
from tensorflow_examples.lite.model_maker.core.task.model_spec.text_spec import mobilebert_qa_spec as MobileBertQaSpec
from tensorflow_examples.lite.model_maker.core.task.model_spec.text_spec import mobilebert_qa_squad_spec as MobileBertQaSquadSpec
from tensorflow_examples.lite.model_maker.core.task.question_answer import create
from tensorflow_examples.lite.model_maker.core.task.question_answer import QuestionAnswer



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Setup for Model Maker.

To build package:
  python setup.py sdist bdist_wheel

To install directly:
  pip install -e .

To uninstall:
  pip uninstall tflite-model-maker
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
import os
import pathlib
import sys
import setup_util

from setuptools import setup

nightly = False
if '--nightly' in sys.argv:
  nightly = True
  sys.argv.remove('--nightly')

project_name = 'tflite-model-maker'
datestring = datetime.datetime.now().strftime('%Y%m%d%H%M')
classifiers = [
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Software Development',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: Libraries :: Python Modules',
]

# Set package version.
if nightly:
  project_name = '{}-nightly'.format(project_name)
  version = '0.4.1'  # Version prefix, usually major.minor.micro.
  version = '{:s}.dev{:s}'.format(version, datestring)
  classifiers += [
      'Development Status :: 4 - Beta',
  ]
else:
  # LINT.IfChange(model_maker_pip_version)
  version = '0.4.0'
  # LINT.ThenChange(../public/__init__.py, ../RELEASE.md)

# Path to model_maker dir: <repo>/tensorflow_examples/lite/model_maker
BASE_DIR = pathlib.Path(os.path.abspath(__file__)).parents[1]
# Path to root dir: <repo>
ROOT_DIR = BASE_DIR.parents[2]
# Original namespace of the lib.
LIB_NAMESPACE = 'tensorflow_examples.lite.model_maker'
# Official package namespace for API. Used as code name.
API_NAMESPACE = 'tflite_model_maker'
# Internal package tflite_model_maker.python mapping internal packages.
INTERNAL_NAME = 'python'
MODEL_MAKER_CONSOLE = 'tflite_model_maker=tflite_model_maker.python.cli.cli:main'

# Build dir `pip_package/src`: copy all source code and create a package.
SRC_NAME = 'src'
BUILD_DIR = BASE_DIR.joinpath('pip_package').joinpath(SRC_NAME)

# Setup options.
setup_options = {
    'package_dir': {
        '': SRC_NAME
    },
    'entry_points': {
        'console_scripts': [MODEL_MAKER_CONSOLE,],
    },
}

DESCRIPTION = ('TFLite Model Maker: a model customization library for on-device'
               ' applications.')
with BASE_DIR.joinpath('README.md').open() as readme_file:
  LONG_DESCRIPTION = readme_file.read()


def _read_required_packages(fpath):
  with fpath.open() as f:
    required_pkgs = [l.strip() for l in f.read().splitlines()]
    required_pkgs = list(
        filter(lambda line: line and not line.startswith('#'), required_pkgs))
  return required_pkgs


def get_required_packages():
  """Gets packages inside requirements.txt."""
  # Gets model maker's required packages
  filename = 'requirements_nightly.txt' if nightly else 'requirements.txt'
  fpath = BASE_DIR.joinpath(filename)
  required_pkgs = _read_required_packages(fpath)

  return required_pkgs


extra_options = setup_util.PackageGen(BASE_DIR, ROOT_DIR, BUILD_DIR, nightly,
                                      version, LIB_NAMESPACE, API_NAMESPACE,
                                      INTERNAL_NAME).run()
setup_options.update(extra_options)

setup(
    name=project_name,
    version=version,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Google LLC',
    author_email='packages@tensorflow.org',
    url='http://github.com/tensorflow/examples',
    download_url='https://github.com/tensorflow/examples/tags',
    license='Apache 2.0',
    scripts=[],
    install_requires=get_required_packages(),
    classifiers=classifiers,
    keywords=['tensorflow', 'lite', 'model customization', 'transfer learning'],
    **setup_options)



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=g-bad-import-order,redefined-builtin
"""Public APIs for TFLite Model Maker, a transfer learning library to train custom TFLite models.

You can install the package with

```bash
pip install tflite-model-maker
```

Typical usage of Model Maker is to create a model in a few lines of code, e.g.:

```python
# Load input data specific to an on-device ML app.
data = DataLoader.from_folder('flower_photos/')
train_data, test_data = data.split(0.9)

# Customize the TensorFlow model.
model = image_classifier.create(train_data)

# Evaluate the model.
accuracy = model.evaluate(test_data)

# Export to Tensorflow Lite model and label file in `export_dir`.
model.export(export_dir='/tmp/')
```

For more details, please refer to our guide:
https://www.tensorflow.org/lite/guide/model_maker.
"""

from tflite_model_maker import audio_classifier
from tflite_model_maker import config
from tflite_model_maker import image_classifier
from tflite_model_maker import model_spec
from tflite_model_maker import object_detector
from tflite_model_maker import question_answer
from tflite_model_maker import recommendation
from tflite_model_maker import searcher
from tflite_model_maker import text_classifier

# Deprecated imports are kept for backward compatiblity and to be removed in
# future versions. Please refer to public APIs for replacement:
# https://www.tensorflow.org/lite/api_docs/python/tflite_model_maker
# pylint: disable=g-bad-import-order
from tensorflow_examples.lite.model_maker.core.data_util.image_dataloader import ImageClassifierDataLoader
from tensorflow_examples.lite.model_maker.core.export_format import ExportFormat
from tensorflow_examples.lite.model_maker.core.task import configs
# pylint: enable=g-bad-import-order

__version__ = '0.4.1.dev202205290509'



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""APIs to train a model that can search Searcher Task."""

import dataclasses
import enum
import logging
import os
import tempfile
from typing import AnyStr, List, Optional

import tensorflow as tf
from tensorflow_examples.lite.model_maker.core.api.api_util import mm_export
import flatbuffers
from tensorflow_lite_support.metadata import metadata_schema_py_generated as _metadata_fb
from tensorflow_lite_support.metadata import schema_py_generated as _schema_fb
from tensorflow_lite_support.metadata.python import metadata as _metadata
from tensorflow_examples.lite.model_maker.core.data_util import searcher_dataloader
from tensorflow_examples.lite.model_maker.core.utils import ondevice_scann_builder
from tensorflow_examples.lite.model_maker.core.utils import scann_converter

# Expected index of the response encoding output tensor in Universal Sentence
# Encoder models.
_USE_RESPONSE_ENCODING_INDEX = 1


@mm_export("searcher.ExportFormat")
@enum.unique
class ExportFormat(enum.Enum):
  TFLITE = "TFLITE"
  SCANN_INDEX_FILE = "SCANN_INDEX_FILE"


@mm_export("searcher.Tree")
@dataclasses.dataclass
class Tree:
  """K-Means partitioning tree configuration.

  In ScaNN, we use single layer K-Means tree to partition the database (index)
  as a way to reduce search space.

  Attributes:
    num_leaves: How many leaves (partitions) to have on the K-Means tree. In
      general, a good starting point would be the square root of the database
      size.
    num_leaves_to_search: During inference ScaNN will compare the query vector
      against all the partition centroids and select the closest
      `num_leaves_to_search` ones to search in. The more leaves to search, the
      better the retrieval quality, and higher computational cost.
    training_sample_size: How many database embeddings to sample for the K-Means
      training. Generally, you want to use a large enough sample of the database
      to train K-Means so that it's representative enough. However, large sample
      can also lead to longer training time. A good starting value would be
      100k, or the whole dataset if it's smaller than that.
    min_partition_size: Smallest allowable cluster size. Any clusters smaller
      than this will be removed, and its data points will be merged with other
      clusters. Recommended to be 1/10 of average cluster size (size of database
      divided by `num_leaves`)
    training_iterations: How many itrations to train K-Means.
    spherical: If true, L2 normalize the K-Means centroids.
    quantize_centroids: If true, quantize centroids to int8.
    random_init: If true, use random init. Otherwise use K-Means++.
  """
  num_leaves: int
  num_leaves_to_search: int
  training_sample_size: int = 100000
  min_partition_size: int = 50
  training_iterations: int = 12
  spherical: bool = False
  quantize_centroids: bool = False
  random_init: bool = True


@mm_export("searcher.ScoreAH")
@dataclasses.dataclass
class ScoreAH:
  """Product Quantization (PQ) based in-partition scoring configuration.

  In ScaNN we use PQ to compress the database embeddings, but not the query
  embedding. We called it Asymmetric Hashing. See
  https://research.google/pubs/pub41694/

  Attributes:
    dimensions_per_block: How many dimensions in each PQ block. If the embedding
      vector dimensionality is a multiple of this value, there will be
      `number_of_dimensions / dimensions_per_block` PQ blocks. Otherwise, the
      last block will be the remainder. For example, if a vector has 12
      dimensions, and `dimensions_per_block` is 2, then there will be 6
      2-dimension blocks. However, if the vector has 13 dimensions and
      `dimensions_per_block` is still 2, there will be 6 2-dimension blocks and
      one 1-dimension block.
    anisotropic_quantization_threshold: If this value is set, we will penalize
      the quantization error that's parallel to the original vector differently
      than the orthogonal error. A generally recommended value for this
      parameter would be 0.2. For more details, please look at ScaNN's 2020 ICML
      paper https://arxiv.org/abs/1908.10396 and the Google AI Blog post
      https://ai.googleblog.com/2020/07/announcing-scann-efficient-vector.html
    training_sample_size: How many database points to sample for training the
      K-Means for PQ centers. A good starting value would be 100k or the whole
      dataset if it's smaller than that.
    training_iterations: How many iterations to run K-Means for PQ.
  """
  dimensions_per_block: int
  anisotropic_quantization_threshold: float = float("nan")
  training_sample_size: int = 100000

  training_iterations: int = 10


@mm_export("searcher.ScoreBruteForce")
@dataclasses.dataclass
class ScoreBruteForce:
  """Bruce force in-partition scoring configuration.

  There'll be no compression or quantization applied to the database
  embeddings or query embeddings.
  """


@mm_export("searcher.ScaNNOptions")
@dataclasses.dataclass
class ScaNNOptions:
  """Options to build ScaNN.

  ScaNN
  (https://ai.googleblog.com/2020/07/announcing-scann-efficient-vector.html) is
  a highly efficient and scalable vector nearest neighbor retrieval
  library from Google Research. We use ScaNN to build the on-device search
  index, and do on-device retrieval with a simplified implementation.
  TODO(b/231134703) Add a link to the README

  Attributes:
    distance_measure: How to compute the distance. Allowed values are
      'dot_product' and 'squared_l2'. Please note that when distance is
      'dot_product', we actually compute the negative dot product between query
      and database vectors, to preserve the notion that "smaller is closer".
    tree: Configure partitioning. If not set, no partitioning is performed.
    score_ah: Configure asymmetric hashing. Must defined this or
      `score_brute_force`.
    score_brute_force: Configure bruce force. Must defined this or `score_ah`.
  """
  distance_measure: str
  tree: Optional[Tree] = None
  score_ah: Optional[ScoreAH] = None
  score_brute_force: Optional[ScoreBruteForce] = None


@mm_export("searcher.Searcher")
class Searcher(object):
  """Creates the similarity search model with ScaNN."""

  def __init__(self,
               serialized_scann_path: str,
               metadata: List[AnyStr],
               embedder_path: Optional[str] = None) -> None:
    """Initializes the Searcher object.

    Args:
      serialized_scann_path: Path to the dir that contains the ScaNN's
        artifacts.
      metadata: The metadata for each of the embeddings in the database. Passed
        in the same order as the embeddings in ScaNN.
      embedder_path: Path to the TFLite Embedder model file.
    """
    self._serialized_scann_path = serialized_scann_path
    self._metadata = metadata
    self._embedder_path = embedder_path

  @classmethod
  def create_from_server_scann(
      cls,
      serialized_scann_path: str,
      metadata: List[AnyStr],
      embedder_path: Optional[str] = None) -> "Searcher":
    """Creates the instance from the serialized serving scann directory.

    Args:
      serialized_scann_path: Path to the dir that contains the ScaNN's
        artifacts.
      metadata: The metadata for each of the embeddings in the database. Passed
        in the same order as the embeddings in ScaNN.
      embedder_path: Path to the TFLite Embedder model file.

    Returns:
      A Searcher instance.
    """
    return cls(serialized_scann_path, metadata, embedder_path)

  @classmethod
  def create_from_data(cls,
                       data: searcher_dataloader.DataLoader,
                       scann_options: ScaNNOptions,
                       cache_dir: Optional[str] = None) -> "Searcher":
    """"Creates the instance from data.

    Args:
      data: Data used to create scann.
      scann_options: Options to build the ScaNN index file.
      cache_dir: The cache directory to save serialized ScaNN and/or the tflite
        model. When cache_dir is not set, a temporary folder will be created and
        will **not** be removed automatically which makes it can be used later.

    Returns:
      A Searcher instance.
    """
    # Gets the ScaNN builder.
    builder = ondevice_scann_builder.builder(
        data.dataset,
        num_neighbors=10,  # This parameter is not used in on-device.
        distance_measure=scann_options.distance_measure)
    if scann_options.tree:
      builder = builder.tree(**dataclasses.asdict(scann_options.tree))
    if scann_options.score_ah:
      # We only support LUT256 for on-device.
      builder = builder.score_ah(
          hash_type="lut256", **dataclasses.asdict(scann_options.score_ah))
    if scann_options.score_brute_force:
      builder = builder.score_brute_force(
          **dataclasses.asdict(scann_options.score_brute_force))

    if cache_dir is None:
      cache_dir = tempfile.mkdtemp()
    if not tf.io.gfile.exists(cache_dir):
      tf.io.gfile.makedirs(cache_dir)
    logging.info("Cache will be stored in %s", cache_dir)

    # Builds, serializes and saves the ScaNN model.
    scann = builder.build()
    serialized_scann_path = os.path.join(cache_dir, "serialized_scann")
    if not tf.io.gfile.exists(serialized_scann_path):
      tf.io.gfile.makedirs(serialized_scann_path)
    scann.serialize(serialized_scann_path)

    return cls(serialized_scann_path, data.metadata, data.embedder_path)

  def export(self,
             export_format: ExportFormat,
             export_filename: str,
             userinfo: AnyStr,
             compression: bool = True):
    """Export the searcher model.

    Args:
      export_format: Export format that could be tflite or on-device ScaNN index
        file, must be `ExportFormat.TFLITE` or `ExportFormat.SCANN_INDEX_FILE`.
      export_filename: File name to save the exported file. The exported file
        can be TFLite model or on-device ScaNN index file.
      userinfo: A special field in the index file that can be an arbitrary
        string supplied by the user.
      compression: Whether to snappy compress the index file.
    """
    export_dir = os.path.dirname(export_filename)
    if not tf.io.gfile.exists(export_dir):
      tf.io.gfile.makedirs(export_dir)

    if export_format is ExportFormat.SCANN_INDEX_FILE:
      output_scann_path = export_filename
    elif export_format is ExportFormat.TFLITE:
      tmpdir = tempfile.mkdtemp()
      output_scann_path = os.path.join(tmpdir, "on_device_scann_index.ldb")
    else:
      raise ValueError("Unsupported export format: ", export_format)

    # Creates the on-device ScaNN index file and saves it in output_scann_path.
    artifacts = scann_converter.convert_serialized_to_on_device(
        self._serialized_scann_path)
    scann_converter.convert_artifacts_to_leveldb(
        output_scann_path,
        metadata=self._metadata,
        artifacts=artifacts,
        userinfo=userinfo,
        compression=compression)

    # Associates the scann index file with the tflite model file.
    if export_format is ExportFormat.TFLITE:
      # Creates the metadata populator.
      if self._embedder_path is None:
        raise ValueError("Can't export the tflite model since embedder model "
                         "is not provided.")
      with tf.io.gfile.GFile(self._embedder_path, "rb") as f:
        model_buffer = f.read()
        populator = _metadata.MetadataPopulator.with_model_buffer(model_buffer)

      # Gets the output index by the number of input tensors.
      model = _schema_fb.Model.GetRootAsModel(model_buffer, 0)
      num_input_tensors = model.Subgraphs(0).InputsLength()

      # Extracts the metadata.
      metadata_buffer = _metadata.get_metadata_buffer(model_buffer)
      if metadata_buffer:
        metadata = _metadata_fb.ModelMetadataT.InitFromObj(
            _metadata_fb.ModelMetadata.GetRootAsModelMetadata(
                metadata_buffer, 0))
      else:
        # Creates the empty metadata.
        input_metadata = [
            _metadata_fb.TensorMetadataT() for i in range(num_input_tensors)
        ]
        num_output_tensors = model.Subgraphs(0).OutputsLength()
        output_metadata = [
            _metadata_fb.TensorMetadataT() for i in range(num_output_tensors)
        ]

        subgraph_metadata = _metadata_fb.SubGraphMetadataT()
        subgraph_metadata.inputTensorMetadata = input_metadata
        subgraph_metadata.outputTensorMetadata = output_metadata

        metadata = _metadata_fb.ModelMetadataT()
        metadata.subgraphMetadata = [subgraph_metadata]

      if num_input_tensors == 3 and (
          not metadata.subgraphMetadata[0].inputProcessUnits):
        # Assume Universal Sentence Encoder-based model.
        output_index = _USE_RESPONSE_ENCODING_INDEX
      else:
        output_index = 0

      # Updates the metadata with the scann associated file.
      scann_file = _metadata_fb.AssociatedFileT()
      scann_file.name = os.path.basename(output_scann_path)
      scann_file.description = "On-device Scann Index file with LevelDB format."
      scann_file.type = _metadata_fb.AssociatedFileType.SCANN_INDEX_FILE
      output_metadata = metadata.subgraphMetadata[0].outputTensorMetadata[
          output_index]
      if output_metadata.associatedFiles is None:
        output_metadata.associatedFiles = [scann_file]
      else:
        output_metadata.associatedFiles.append(scann_file)

      # Saves the updated metadata and the scann associated file along with the
      # model buffer.
      buffer_builder = flatbuffers.Builder(0)
      buffer_builder.Finish(
          metadata.Pack(buffer_builder),
          _metadata.MetadataPopulator.METADATA_FILE_IDENTIFIER)
      updated_metadata_buffer = buffer_builder.Output()
      populator.load_metadata_buffer(updated_metadata_buffer)
      populator.load_associated_files([output_scann_path])
      populator.populate()

      output_tflite_path = export_filename
      with tf.io.gfile.GFile(output_tflite_path, "wb") as f:
        f.write(populator.get_model_buffer())
      tf.io.gfile.rmtree(tmpdir)



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ScannBuilder class for on-device applications."""

from google.protobuf import text_format
from scann.proto import scann_pb2
from scann.scann_ops.py import scann_builder
from scann.scann_ops.py import scann_ops_pybind


def builder(db, num_neighbors, distance_measure):
  """pybind analogue of builder() in scann_ops.py for the on-device use case."""

  def builder_lambda(db, config, training_threads, **kwargs):
    return scann_ops_pybind.create_searcher(db, config, training_threads,
                                           **kwargs)

  return OndeviceScannBuilder(
      db, num_neighbors, distance_measure).set_builder_lambda(builder_lambda)


class OndeviceScannBuilder(scann_builder.ScannBuilder):
  """ScannBuilder for on-device applications."""

  def create_config(self):
    """Creates the config."""
    config = super().create_config()
    config_proto = scann_pb2.ScannConfig()
    text_format.Parse(config, config_proto)
    # We don't support residual quantization on device so we need to disable
    # use_residual_quantization.
    if config_proto.hash.asymmetric_hash.use_residual_quantization:
      config_proto.hash.asymmetric_hash.use_residual_quantization = False
    return text_format.MessageToString(config_proto)
